const {
    time,
    loadFixture,
  } = require("@nomicfoundation/hardhat-network-helpers");
  const fs = require('fs')
    const path = require('path')
  const { anyValue } = require("@nomicfoundation/hardhat-chai-matchers/withArgs");
  const { expect } = require("chai");
  const { ethers, network } = require("hardhat");
  const assert = require("assert");
  const { StandardMerkleTree } = require('@openzeppelin/merkle-tree');
  const { getContractAddress } = require('@ethersproject/address')
  const createPermit = require('./utils/createPermit')
  const { keccak256 } = require('@ethersproject/keccak256');
  ((network.name != 'multipleNetworks') ? describe : describe.skip)("eeseeOffchain", function () {
    let ESE;
    let ERC20;
    let eesee;
    let NFT;
    let ERC1155
    let signer, acc2, acc3, acc4, acc5, acc6, acc7, acc8, acc9, feeCollector;
    let ticketBuyers;
    let minter;
    let staking
    let accessManager
    let mockStorage

    const zeroAddress = "0x0000000000000000000000000000000000000000"
    const oneAddress = "0x0000000000000000000000000000000000000001"
    let snapshotId
    this.beforeAll(async() => {
        [signer, acc2, acc3, acc4, acc5, acc6, acc7, acc8, acc9, feeCollector, royaltyCollector] = await ethers.getSigners()
        ticketBuyers = [acc2,acc3, acc4, acc5, acc6,  acc7]
        const _ESE = await hre.ethers.getContractFactory("ESE");
        const _MockERC20 = await hre.ethers.getContractFactory("MockERC20");
        const _MockStorage = await hre.ethers.getContractFactory("MockStorage");
        const _eesee = await hre.ethers.getContractFactory("EeseeOffchain");
        const _NFT = await hre.ethers.getContractFactory("EeseeNFT");
        const _NFTDrop = await hre.ethers.getContractFactory("EeseeNFTDrop");
        const _NFTlazyMint = await hre.ethers.getContractFactory("EeseeNFTLazyMint");
        const _minter = await hre.ethers.getContractFactory("EeseeMinter");

        const _eeseeStaking = await hre.ethers.getContractFactory("EeseeStaking");
        const _eeseeAccessManager = await hre.ethers.getContractFactory("EeseeAccessManager");
        const _eeseeProxy = await hre.ethers.getContractFactory("EeseeProxy");

        mockStorage = await _MockStorage.deploy()
        await mockStorage.deployed()

        ESE = await _ESE.deploy([{
                cliff: 0,
                duration: 0,
                TGEMintShare: 10000
            }
        ])
        await ESE.deployed()
        await ESE.addVestingBeneficiaries(0, [{addr: signer.address, amount: '2000000000000000000000000'}])
        await ESE.initialize(0)

        ERC20 = await _MockERC20.deploy('20000000000000000000000000000')
        await ERC20.deployed()

        accessManager = await _eeseeAccessManager.deploy();
        await accessManager.deployed()

        // SIGNER_ROLE
        await accessManager.grantRole('0xe2f4eaae4a9751e85a3e4a7b9587827a877f29914755229b07a7b2da98285f70', acc8.address)

        const NFTDrop = await _NFTDrop.deploy(zeroAddress)
        const NFTLazyMint = await _NFTlazyMint.deploy(zeroAddress)
        minter = await _minter.deploy(NFTLazyMint.address, NFTDrop.address, oneAddress)
        await minter.deployed()

        proxy = await _eeseeProxy.deploy();
        await proxy.deployed()

        staking = await _eeseeStaking.deploy(
            ESE.address, 
            [{volumeBreakpoint: 500, rewardRateFlexible: 500000, rewardRateLocked: 500000}], 
            accessManager.address,
            zeroAddress
        )
        await staking.deployed()

        eesee = await _eesee.deploy(
            ESE.address, 
            staking.address,
            proxy.address,
            minter.address, 
            accessManager.address,
            zeroAddress,
            zeroAddress,
            zeroAddress
        )
        await eesee.deployed()

        await staking.grantVolumeUpdater(eesee.address)

        NFT = await _NFT.deploy()
        await NFT.initialize({
            name: "APES",
            symbol:"bayc",
            baseURI: "/",
            revealedURI: "",
            contractURI:'/',
            royaltyReceiver: zeroAddress,
            royaltyFeeNumerator: 0
        },10, signer.address)

        await NFT.deployed()
        await NFT.setApprovalForAll(eesee.address, true)
        await NFT.connect(acc4).setApprovalForAll(eesee.address, true)

        for (let i = 0; i < ticketBuyers.length; i++) {
            await ESE.transfer(ticketBuyers[i].address, '10000000000000000000000')
            await ESE.connect(ticketBuyers[i]).approve(eesee.address, '10000000000000000000000')
        }

        await ESE.transfer(acc8.address, '100000000000000000000')
        await ESE.connect(acc8).approve(eesee.address, '100000000000000000000')

        const _ERC1155 = await hre.ethers.getContractFactory("Mock1155");
        ERC1155 = await _ERC1155.deploy("", 1, 10000, royaltyCollector.address, 10000)
        await ERC1155.deployed()
    })

    let nonce = 0
    const createLot = async (asset, data, deadline, _signer) => {
        const types = {
            CreateLot: [
              {
                name: "assetHash",
                type: "bytes32"
              },
              {
                name: "dataHash",
                type: "bytes32"
              },
              {
                name: "nonce",
                type: "uint256"
              },
              {
                name: "deadline",
                type: "uint256"
              },
            ],
          }
        const eip712Domain = await eesee.eip712Domain()
        const domain = {
            name: eip712Domain.name,
            version: eip712Domain.version,
            chainId: eip712Domain.chainId,
            verifyingContract: eip712Domain.verifyingContract
        }
        const abi = ethers.utils.defaultAbiCoder;
        const assetEncoded = abi.encode(
            ["tuple(address, uint256, uint256, uint8, bytes)"], 
            [[asset.token, asset.tokenID, asset.amount, asset.assetType, asset.data]]
        );

        const values = {assetHash: keccak256(assetEncoded), dataHash: keccak256(data), nonce,deadline}
        let signature = await _signer._signTypedData(domain, types, values)

        const params = {nonce, deadline, signature}
        nonce += 1
        return params
    }

    const buyTickets = async (ID, amount, recipient, deadline, _signer) => {
      const types = {
          BuyTickets: [
            {
              name: "ID",
              type: "bytes32"
            },
            {
              name: "amount",
              type: "uint256"
            },
            {
              name: "recipient",
              type: "address"
            },
            {
              name: "nonce",
              type: "uint256"
            },
            {
              name: "deadline",
              type: "uint256"
            },
          ],
        }
      const eip712Domain = await eesee.eip712Domain()
      const domain = {
          name: eip712Domain.name,
          version: eip712Domain.version,
          chainId: eip712Domain.chainId,
          verifyingContract: eip712Domain.verifyingContract
      }

      const values = {ID, amount, recipient, nonce, deadline}
      let signature = await _signer._signTypedData(domain, types, values)

      const params = {nonce, deadline, signature}
      nonce += 1
      return params
    }

    const claimAssets = async (ID, asset, sender, deadline, _signer) => {
      const types = {
        ClaimAsset: [
            {
              name: "ID",
              type: "bytes32"
            },
            {
              name: "assetHash",
              type: "bytes32"
            },
            {
              name: "sender",
              type: "address"
            },
            {
              name: "nonce",
              type: "uint256"
            },
            {
              name: "deadline",
              type: "uint256"
            },
          ],
        }
      const eip712Domain = await eesee.eip712Domain()
      const domain = {
          name: eip712Domain.name,
          version: eip712Domain.version,
          chainId: eip712Domain.chainId,
          verifyingContract: eip712Domain.verifyingContract
      }
      const abi = ethers.utils.defaultAbiCoder;
      const assetEncoded = abi.encode(
          ["tuple(address, uint256, uint256, uint8, bytes)"], 
          [[asset.token, asset.tokenID, asset.amount, asset.assetType, asset.data]]
      );

      const values = {ID, assetHash: keccak256(assetEncoded), sender, nonce, deadline}
      let signature = await _signer._signTypedData(domain, types, values)

      const params = {nonce, deadline, signature}
      nonce += 1
      return params
    }

    const claimTokens = async (ID, amount, sender, deadline, _signer) => {
      const types = {
        ClaimTokens: [
            {
              name: "ID",
              type: "bytes32"
            },
            {
              name: "amount",
              type: "uint256"
            },
            {
              name: "sender",
              type: "address"
            },
            {
              name: "nonce",
              type: "uint256"
            },
            {
              name: "deadline",
              type: "uint256"
            },
          ],
        }
      const eip712Domain = await eesee.eip712Domain()
      const domain = {
          name: eip712Domain.name,
          version: eip712Domain.version,
          chainId: eip712Domain.chainId,
          verifyingContract: eip712Domain.verifyingContract
      }

      const values = {ID, amount, sender, nonce, deadline}
      let signature = await _signer._signTypedData(domain, types, values)

      const params = {nonce, deadline, signature}
      nonce += 1
      return params
    }

    let id1
    let id2

    it('Lists NFT', async () => {
        const asset1 = {token: NFT.address, tokenID: 1, amount: 1, assetType: 0, data:'0x'}
        const asset2 = {token: NFT.address, tokenID: 2, amount: 1, assetType: 0, data:'0x'}
        const sig1 = await createLot(asset1, "0x01", '999999999999', acc8)
        const sig2 = await createLot(asset2, "0x02", '999999999999', acc8)
        const sig3 = await createLot(asset2, "0x02", '999999999999', acc7)
        const sig4 = await createLot(asset2, "0x02", '2', acc8)

        id1 = await eesee.getCreateLotStructHash(asset1, "0x01", 0, '999999999999')
        id2 = await eesee.getCreateLotStructHash(asset2, "0x02", 1, '999999999999')
        await expect(eesee.connect(signer).createLots(
            [
                asset1, 
                asset2
            ], 
            ["0x01", "0x02"],
            []
        )).to.be.revertedWithCustomError(eesee, "InvalidArrayLengths")

        await expect(eesee.connect(signer).createLots(
          [
              asset1, 
              asset2
          ], 
          ["0x01"],
          [sig1, sig2]
      )).to.be.revertedWithCustomError(eesee, "InvalidArrayLengths")

        await expect(eesee.connect(signer).createLots(
          [
              asset1, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig1, sig3]
        )).to.be.revertedWithCustomError(eesee, "InvalidSignature")

        await expect(eesee.connect(signer).createLots(
          [
              asset1, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig1, sig4]
        )).to.be.revertedWithCustomError(eesee, "ExpiredDeadline")

        const sig5 = await createLot({...asset1, amount: 0}, "0x01", '9999999999999', acc8)
        await expect(eesee.connect(signer).createLots(
          [
              {...asset1, amount: 0}, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig5, sig2]
        )).to.be.revertedWithCustomError(eesee, "InvalidAmount")

        const sig6 = await createLot({...asset1, token: ERC1155.address}, "0x01", '9999999999999', acc8)
        await expect(eesee.connect(signer).createLots(
          [
              {...asset1, token: ERC1155.address}, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig6, sig2]
        )).to.be.revertedWithCustomError(eesee, "InvalidInterface")

        const sig7 = await createLot({...asset1, token: staking.address}, "0x01", '9999999999999', acc8)
        await expect(eesee.connect(signer).createLots(
          [
              {...asset1, token: staking.address}, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig7, sig2]
        )).to.be.revertedWithCustomError(eesee, "InvalidInterface")

        await expect(eesee.connect(signer).createLots(
          [
              asset1, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig1, sig2]
        )).to.emit(eesee, "CreateLot").withArgs(id1, signer.address, anyValue, "0x01", 0)
        .and.to.emit(eesee, "CreateLot").withArgs(id2, signer.address, anyValue, "0x02", 1)

        await expect(eesee.connect(signer).createLots(
          [
              asset1, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig1, sig2]
        )).to.be.revertedWithCustomError(eesee, "NonceUsed")
    })

    it('Creates lots with ERC20 asset types', async () => {
      await ERC20.approve(eesee.address, 10000)

      const asset1 = {token: ERC20.address, tokenID: 1, amount: 10, assetType: 2, data:'0x'}
      const asset2 = {token: ERC20.address, tokenID: 1, amount: 100, assetType: 2, data:'0x'}
      const nonce1 = nonce
      const sig1 = await createLot(asset1, "0x01", '999999999999', acc8)
      const nonce2 = nonce
      const sig2 = await createLot(asset2, "0x02", '999999999999', acc8)

      const _id1 = await eesee.getCreateLotStructHash(asset1, "0x01", nonce1, '999999999999')
      const _id2 = await eesee.getCreateLotStructHash(asset2, "0x02", nonce2, '999999999999')

      const sig5 = await createLot({...asset1, amount: 0}, "0x01", '9999999999999', acc8)
      await expect(eesee.connect(signer).createLots(
        [
            {...asset1, amount: 0}, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig5, sig2]
      )).to.be.revertedWithCustomError(eesee, "InvalidAmount")

      const sig6 = await createLot({...asset1, token: NFT.address}, "0x01", '9999999999999', acc8)
      await expect(eesee.connect(signer).createLots(
        [
            {...asset1, token: NFT.address}, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig6, sig2]
      )).to.be.revertedWithCustomError(eesee, "InvalidInterface")

      await expect(eesee.connect(signer).createLots(
        [
            asset1, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig1, sig2]
      )).to.emit(eesee, "CreateLot").withArgs(_id1, signer.address, anyValue, "0x01", nonce1)
      .and.to.emit(eesee, "CreateLot").withArgs(_id2, signer.address, anyValue, "0x02", nonce2)     
    })

    it('Creates lots with ERC1155 asset types', async () => {
        await ERC1155.setApprovalForAll(eesee.address, 10000)

        const asset1 = {token: ERC1155.address, tokenID: 1, amount: 10, assetType: 1, data:'0x'}
        const asset2 = {token: ERC1155.address, tokenID: 1, amount: 100, assetType: 1, data:'0x'}
        const nonce1 = nonce
        const sig1 = await createLot(asset1, "0x01", '999999999999', acc8)
        const nonce2 = nonce
        const sig2 = await createLot(asset2, "0x02", '999999999999', acc8)

        const _id1 = await eesee.getCreateLotStructHash(asset1, "0x01", nonce1, '999999999999')
        const _id2 = await eesee.getCreateLotStructHash(asset2, "0x02", nonce2, '999999999999')

        const sig5 = await createLot({...asset1, amount: 0}, "0x01", '9999999999999', acc8)
        await expect(eesee.connect(signer).createLots(
          [
              {...asset1, amount: 0}, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig5, sig2]
        )).to.be.revertedWithCustomError(eesee, "InvalidAmount")

        const sig6 = await createLot({...asset1, token: NFT.address}, "0x01", '9999999999999', acc8)
        await expect(eesee.connect(signer).createLots(
          [
              {...asset1, token: NFT.address}, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig6, sig2]
        )).to.be.revertedWithCustomError(eesee, "InvalidInterface")

        const sig7 = await createLot({...asset1, token: staking.address}, "0x01", '9999999999999', acc8)
        await expect(eesee.connect(signer).createLots(
          [
              {...asset1, token: staking.address}, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig7, sig2]
        )).to.be.revertedWithCustomError(eesee, "InvalidInterface")

        await expect(eesee.connect(signer).createLots(
          [
              asset1, 
              asset2
          ], 
          ["0x01", "0x02"],
          [sig1, sig2]
        )).to.emit(eesee, "CreateLot").withArgs(_id1, signer.address, anyValue, "0x01", nonce1)
        .and.to.emit(eesee, "CreateLot").withArgs(_id2, signer.address, anyValue, "0x02", nonce2)     
    })

    it('Creates lots with native asset types', async () => {
      const asset1 = {token: zeroAddress, tokenID: 1, amount: 10, assetType: 3, data:'0x'}
      const asset2 = {token: ERC20.address, tokenID: 10, amount: 100, assetType: 3, data:'0x'}
      const nonce1 = nonce
      const sig1 = await createLot(asset1, "0x01", '999999999999', acc8)
      const nonce2 = nonce
      const sig2 = await createLot(asset2, "0x02", '999999999999', acc8)

      const _id1 = await eesee.getCreateLotStructHash(asset1, "0x01", nonce1, '999999999999')
      const _id2 = await eesee.getCreateLotStructHash(asset2, "0x02", nonce2, '999999999999')

      const sig5 = await createLot({...asset1, amount: 0}, "0x01", '9999999999999', acc8)
      await expect(eesee.connect(signer).createLots(
        [
            {...asset1, amount: 0}, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig5, sig2],
        {value: 110}
      )).to.be.revertedWithCustomError(eesee, "InvalidAmount")

      await expect(eesee.connect(signer).createLots(
        [
            asset1, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig1, sig2],
        {value: 109}
      )).to.be.revertedWithCustomError(eesee, "InvalidValue")

      await expect(eesee.connect(signer).createLots(
        [
            asset1, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig1, sig2],
        {value: 111}
      )).to.be.revertedWithCustomError(eesee, "InvalidValue")

      await expect(eesee.connect(signer).createLots(
        [
            asset1, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig1, sig2],
        {value: 110}
      )).to.emit(eesee, "CreateLot").withArgs(_id1, signer.address, anyValue, "0x01", nonce1)
      .and.to.emit(eesee, "CreateLot").withArgs(_id2, signer.address, anyValue, "0x02", nonce2) 
    })

    it('ESE type lot', async () => {
      const asset1 = {token: zeroAddress, tokenID: 1, amount: 500, assetType: 4, data:'0x'}
      const asset2 = {token: ERC20.address, tokenID: 10, amount: 1000, assetType: 4, data:'0x'}
      const nonce1 = nonce
      const sig1 = await createLot(asset1, "0x01", '999999999999', acc8)
      const nonce2 = nonce
      const sig2 = await createLot(asset2, "0x02", '999999999999', acc8)

      const _id1 = await eesee.getCreateLotStructHash(asset1, "0x01", nonce1, '999999999999')
      const _id2 = await eesee.getCreateLotStructHash(asset2, "0x02", nonce2, '999999999999')

      await expect(eesee.connect(signer).createLots(
        [
            asset1, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig1, sig2]
      )).to.emit(eesee, "CreateLot").withArgs(_id1, signer.address, anyValue, "0x01", nonce1)
      .and.to.emit(eesee, "CreateLot").withArgs(_id2, signer.address, anyValue, "0x02", nonce2) 
    })

    it('ERC721LazyMint type lot', async () => {
      const asset1 = {token: zeroAddress, tokenID: 1, amount: 1, assetType: 5, data:'0x'}
      const asset2 = {token: ERC20.address, tokenID: 10, amount: 1, assetType: 5, data:'0x'}
      const nonce1 = nonce
      const sig1 = await createLot(asset1, "0x01", '999999999999', acc8)
      const nonce2 = nonce
      const sig2 = await createLot(asset2, "0x02", '999999999999', acc8)

      const _id1 = await eesee.getCreateLotStructHash(asset1, "0x01", nonce1, '999999999999')
      const _id2 = await eesee.getCreateLotStructHash(asset2, "0x02", nonce2, '999999999999')

      const sig5 = await createLot({...asset1, amount: 0}, "0x01", '9999999999999', acc8)
      await expect(eesee.connect(signer).createLots(
        [
            {...asset1, amount: 0}, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig5, sig2]
      )).to.be.revertedWithCustomError(eesee, "InvalidAmount")

      await expect(eesee.connect(signer).createLots(
        [
            asset1, 
            asset2
        ], 
        ["0x01", "0x02"],
        [sig1, sig2]
      )).to.emit(eesee, "CreateLot").withArgs(_id1, signer.address, anyValue, "0x01", nonce1)
      .and.to.emit(eesee, "CreateLot").withArgs(_id2, signer.address, anyValue, "0x02", nonce2) 
    })

    it('Buys tickets with permit', async () => {
      await ESE.connect(acc2).approve(eesee.address, 0)
      const currentTimestamp = (await ethers.provider.getBlock()).timestamp
      const deadline = currentTimestamp + 10000
      const correctPermit = await createPermit(acc2, eesee, '40', deadline, ESE)

      const params_ = ethers.utils.defaultAbiCoder.encode(
        ["uint256", "uint256", "uint8", "bytes32", "bytes32"],
        ['40', deadline, correctPermit.v, correctPermit.r, correctPermit.s]
      );

      const nonce1 = nonce
      const sig1 = await buyTickets(id1, 10, acc2.address, deadline, acc8)
      const nonce2 = nonce
      const sig2 = await buyTickets(id2, 20, acc2.address, deadline, acc8)

      await expect(eesee.connect(acc2).buyTickets([id1, id2], [10, 20], [sig1], zeroAddress, params_)).to.be.revertedWithCustomError(eesee, "InvalidRecipient")
      await expect(eesee.connect(acc2).buyTickets([id1, id2], [10, 20], [sig1], acc2.address, params_)).to.be.revertedWithCustomError(eesee, "InvalidArrayLengths")
      await expect(eesee.connect(acc2).buyTickets([id1, id2], [10], [sig1, sig2], acc2.address, params_)).to.be.revertedWithCustomError(eesee, "InvalidArrayLengths")
      const balanceBefore = await ESE.balanceOf(acc2.address)
      const incorrectPermit = await createPermit(acc2, eesee, '10', deadline, ESE)
      const params_incorrect = ethers.utils.defaultAbiCoder.encode(
          ["uint256", "uint256", "uint8", "bytes32", "bytes32"],
          ['10', deadline, incorrectPermit.v, incorrectPermit.r, incorrectPermit.s]
        );

        const params_incorrect2 = ethers.utils.defaultAbiCoder.encode(
          ["uint256", "uint256", "uint8", "bytes32", "bytes32"],
          ['11', deadline, incorrectPermit.v, incorrectPermit.r, incorrectPermit.s]
        );

      await expect(eesee.connect(acc2).buyTickets([id1, id2], [10, 20], [sig1, sig2], acc2.address, params_incorrect2))
      .to.be.revertedWith("ERC20Permit: invalid signature")
      await expect(eesee.connect(acc2).buyTickets([id1, id2], [10, 20], [sig1, sig2], acc2.address, params_incorrect))
      .to.be.revertedWith("ERC20: insufficient allowance")

      await expect(eesee.connect(acc2).buyTickets([id1, id2], [10, 20], [sig1, sig2], acc2.address, params_))
          .to.emit(eesee, "BuyTickets").withArgs(id1, acc2.address, acc2.address, 10, nonce1)
          .and.to.emit(eesee, "BuyTickets").withArgs(id2, acc2.address, acc2.address, 20, nonce2)


      const balanceAfter = await ESE.balanceOf(acc2.address)
      if(network.name != 'testnet') {
          // Hardhat node has trouble with try catch statements, so we exclude this check for 'testnet'
          // I tested in on the live goerli chain and it was OK.
          assert.equal(await staking.volume(acc2.address), 10 + 20, "Volume is correct")
      }

      assert.equal(BigInt(balanceBefore) - BigInt(balanceAfter), 10 + 20, "Price paid is correct")

      await expect(eesee.connect(acc8).buyTickets([id1], [0], [{nonce: 9999, deadline: "999999999999", signature: "0x"}], acc2.address, "0x"))
      .to.emit(eesee, "BuyTickets").withArgs(id1, acc8.address, acc2.address, 0, 9999)
    })

    it('Buys tickets after volumeUpdater rights were removed from eesee', async () => {
      await ESE.connect(acc2).approve(eesee.address, 500)

      const _nonce = nonce
      const sig = await buyTickets(id1, 500, acc3.address, '999999999999', acc8)

      const volumeBefore = await staking.volume(acc2.address)

      await staking.revokeVolumeUpdater(eesee.address)
      await expect(eesee.connect(acc2).buyTickets([id1], [500], [sig], acc3.address, "0x"))
          .to.emit(eesee, "BuyTickets").withArgs(id1, acc2.address, acc3.address, 500, _nonce)
          .and.to.emit(eesee, "AddVolumeReverted").withArgs(500, acc3.address)

      const volumeAfter = await staking.volume(acc2.address)

      assert.equal(volumeBefore.toString(), volumeAfter.toString(), "Volume is correct")
    })

    it('Claims asset', async () => {
      const asset1 = {token: NFT.address, tokenID: 1, amount: 1, assetType: 0, data:'0x'}
      const asset2 = {token: NFT.address, tokenID: 2, amount: 1, assetType: 0, data:'0x'}
      const nonce1 = nonce
      const sig1 = await claimAssets(id1, asset1, acc3.address, '999999999999', acc8)
      const nonce2 = nonce
      const sig2 = await claimAssets(id2, asset2, acc3.address, '999999999999', acc8)

      await expect(eesee.connect(signer).claimAssets(
        [id1, id2],
        [
          asset1, 
          asset2
        ], 
        [sig1],
        zeroAddress,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "InvalidRecipient")

      await expect(eesee.connect(signer).claimAssets(
        [id1, id2],
        [
          asset1, 
          asset2
        ], 
        [sig1],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "InvalidArrayLengths")

      await expect(eesee.connect(signer).claimAssets(
        [id1, id2],
        [
          asset1
        ], 
        [sig1, sig2],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "InvalidArrayLengths")

      await expect(eesee.connect(acc4).claimAssets(
        [id1, id2],
        [
          asset1, 
          asset2
        ], 
        [sig1, sig2],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "InvalidSignature")

      await expect(eesee.connect(acc3).claimAssets(
        [id1, id2],
        [
          asset1, 
          asset2
        ], 
        [sig1, sig2],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      ))
      .to.emit(eesee, "ClaimAsset").withArgs(id1, acc3.address, acc4.address, anyValue, nonce1)
      .and.to.emit(eesee, "ClaimAsset").withArgs(id2, acc3.address, acc4.address, anyValue, nonce2)

      assert.equal(await NFT.ownerOf(1), acc4.address, "Asset not transfered")
      assert.equal(await NFT.ownerOf(2), acc4.address, "Asset not transfered")


      await NFT.transferFrom(signer.address, eesee.address, 4)
      let asset = {token: NFT.address, tokenID: 4, amount: 1, assetType: 0, data:'0x'}
      const storeData = mockStorage.interface.encodeFunctionData('storeString', ["Hello World"])
      await expect(eesee.connect(acc8).claimAssets(
        [id1],
        [asset], 
        [{nonce: 999, deadline: '999999999999', signature: "0x"}],
        acc4.address,
        {target: mockStorage.address, callData: storeData}
      ))
      .to.emit(eesee, "ClaimAsset").withArgs(id1, acc8.address, acc4.address, anyValue, 999)

      assert.equal(await NFT.ownerOf(4), acc4.address, "Asset not transfered")
      assert.equal(await mockStorage.newStr(), "Hello World", "Call not made")

      asset = {token: ERC1155.address, tokenID: 1, amount: 13, assetType: 1, data:'0x'}
      await expect(eesee.connect(acc8).claimAssets(
        [id1],
        [asset], 
        [{nonce: 1001, deadline: '999999999999', signature: "0x"}],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      ))
      .to.emit(eesee, "ClaimAsset").withArgs(id1, acc8.address, acc4.address, anyValue, 1001)
      assert.equal(await ERC1155.balanceOf(acc4.address, 1), 13, "Asset not transfered")

      asset = {token: ERC20.address, tokenID: 1, amount: 13, assetType: 2, data:'0x'}
      await expect(eesee.connect(acc8).claimAssets(
        [id1],
        [asset], 
        [{nonce: 1002, deadline: '999999999999', signature: "0x"}],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      ))
      .to.emit(eesee, "ClaimAsset").withArgs(id1, acc8.address, acc4.address, anyValue, 1002)
      assert.equal(await ERC20.balanceOf(acc4.address), 13, "Asset not transfered")

      const _balanceBefore = await ethers.provider.getBalance(acc4.address);
      asset = {token: zeroAddress, tokenID: 1, amount: 13, assetType: 3, data:'0x'}
      await expect(eesee.connect(acc8).claimAssets(
        [id1],
        [asset], 
        [{nonce: 1003, deadline: '999999999999', signature: "0x"}],
        eesee.address,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "TransferNotSuccessful")

      await expect(eesee.connect(acc8).claimAssets(
        [id1],
        [asset], 
        [{nonce: 1003, deadline: '999999999999', signature: "0x"}],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      ))
      .to.emit(eesee, "ClaimAsset").withArgs(id1, acc8.address, acc4.address, anyValue, 1003)
      const _balanceAfter =  await ethers.provider.getBalance(acc4.address);
      assert.equal(_balanceAfter.sub(_balanceBefore).toString(), "13", "Asset not transfered")

      const balanceBefore = await ESE.balanceOf(acc4.address)
      asset = {token: ERC20.address, tokenID: 1, amount: 13, assetType: 4, data:'0x'}
      await expect(eesee.connect(acc8).claimAssets(
        [id1],
        [asset], 
        [{nonce: 1004, deadline: '999999999999', signature: "0x"}],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      ))
      .to.emit(eesee, "ClaimAsset").withArgs(id1, acc8.address, acc4.address, anyValue, 1004)
      const balanceAfter = await ESE.balanceOf(acc4.address)
      assert.equal(BigInt(balanceAfter) - BigInt(balanceBefore), 13, "Claim is correct")

      const data = ethers.utils.defaultAbiCoder.encode(
            [
                "uint256", 
                "address", 
                {
                    type: "tuple",
                    name: "collectionMetadata",
                    components: [
                      { name: "name", type: 'string' },
                      { name: "symbol", type: 'string' },
                      { name: "contractURI", type:'string' },
                    ]
                },
                {
                    type: "tuple",
                    name: "tokenMetadata",
                    components: [
                      { name: "URI", type:'string' },
                      { name: "royaltyReceiver", type:'address' },
                      { name: "royaltyFeeNumerator", type:'uint96' },
                    ]
                }
            ],
            [777, acc2.address, {name:"hello", symbol:"HI", contractURI:"/contract"}, {URI: "/hello", royaltyReceiver:acc3.address, royaltyFeeNumerator:100},]
          );
      asset = {token: ERC20.address, tokenID: 1, amount: 13, assetType: 5, data:data}
      await expect(eesee.connect(acc8).claimAssets(
        [id1],
        [asset], 
        [{nonce: 1005, deadline: '999999999999', signature: "0x"}],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      ))
      .to.emit(eesee, "ClaimAsset").withArgs(id1, acc8.address, acc4.address, anyValue, 1005)
      .to.emit(minter, "LazyMint").withArgs(anyValue, 1, acc4.address)
    })

    it('Claims tokens', async () => {
      const nonce1 = nonce
      const sig1 = await claimTokens(id1, 13, acc3.address, '999999999999', acc8)
      const nonce2 = nonce
      const sig2 = await claimTokens(id2, 17, acc3.address, '999999999999', acc8)

      const balanceBefore = await ESE.balanceOf(acc4.address)
      await expect(eesee.connect(acc3).claimTokens(
        [id1, id2],
        [
          13, 
          17
        ], 
        [sig1],
        zeroAddress,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "InvalidRecipient")

      await expect(eesee.connect(acc3).claimTokens(
        [id1, id2],
        [
          13, 
          17
        ], 
        [sig1],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "InvalidArrayLengths")

      await expect(eesee.connect(acc3).claimTokens(
        [id1, id2],
        [13], 
        [sig1, sig2],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "InvalidArrayLengths")


      await expect(eesee.connect(acc3).claimTokens(
        [id1, id2],
        [
          14, 
          17
        ], 
        [sig1, sig2],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      )).to.be.revertedWithCustomError(eesee, "InvalidSignature")

      await expect(eesee.connect(acc3).claimTokens(
        [id1, id2],
        [
          13, 
          17
        ], 
        [sig1, sig2],
        acc4.address,
        {target: zeroAddress, callData: "0x"}
      ))
      .to.emit(eesee, "ClaimTokens").withArgs(id1, acc3.address, acc4.address, 13, nonce1)
      .and.to.emit(eesee, "ClaimTokens").withArgs(id2, acc3.address, acc4.address, 17, nonce2)

      const balanceAfter = await ESE.balanceOf(acc4.address)
      assert.equal(BigInt(balanceAfter) - BigInt(balanceBefore), 13 + 17, "Claim is correct")

      const storeData = mockStorage.interface.encodeFunctionData('storeString', ["Hello World2"])
      await expect(eesee.connect(acc8).claimTokens(
        [id1],
        [100], 
        [{nonce: 1000, deadline: '999999999999', signature: "0x"}],
        acc4.address,
        {target: mockStorage.address, callData: storeData}
      ))
      .to.emit(eesee, "ClaimTokens").withArgs(id1, acc8.address, acc4.address, 100, 1000)

      const balanceAfterAfter = await ESE.balanceOf(acc4.address)
      assert.equal(BigInt(balanceAfterAfter) - BigInt(balanceAfter), 100, "Claim is correct")
      assert.equal(await mockStorage.newStr(), "Hello World2", "Call not made")
    })

    it('revokes signature', async () => {
      const asset1 = {token: NFT.address, tokenID: 1, amount: 1, assetType: 0, data:'0x'}
      const _nonce1 = nonce
      const sig1 = await createLot(asset1, "0x01", '999999999999', acc8)

      const asset2 = {token: NFT.address, tokenID: 2, amount: 1, assetType: 0, data:'0x'}
      const _nonce2 = nonce
      const sig2 = await createLot(asset2, "0x02", '999999999999', acc8)

      await expect(eesee.connect(acc4).revokeSignatures([_nonce1, _nonce2])).to.be.revertedWithCustomError(eesee, "CallerNotAuthorized")
      await expect(eesee.connect(acc8).revokeSignatures([_nonce1, _nonce2]))
      .to.emit(eesee, "RevokeSignature").withArgs(_nonce1)
      .and.to.emit(eesee, "RevokeSignature").withArgs(_nonce2)

      await expect(eesee.connect(acc4).createLots(
        [asset1], 
        ["0x01"],
        [sig1]
      )).to.be.revertedWithCustomError(eesee, "NonceUsed")

      await expect(eesee.connect(acc4).createLots(
        [asset2], 
        ["0x02"],
        [sig2]
      )).to.be.revertedWithCustomError(eesee, "NonceUsed")

      await expect(eesee.connect(acc8).revokeSignatures([_nonce1]))
      .to.not.emit(eesee, "RevokeSignature")
    })

    it('Reverts constructor', async () => {
      const _eesee = await hre.ethers.getContractFactory("EeseeOffchain");

      await expect(_eesee.deploy(
          zeroAddress, 
          staking.address,
          proxy.address,
          minter.address, 
          accessManager.address,
          zeroAddress,
          zeroAddress,
          zeroAddress
      )).to.be.revertedWithCustomError(_eesee, "InvalidConstructor")

      await expect(_eesee.deploy(
          ESE.address, 
          staking.address,
          zeroAddress,
          minter.address, 
          accessManager.address,
          zeroAddress,
          zeroAddress,
          zeroAddress
      )).to.be.revertedWithCustomError(_eesee, "InvalidConstructor")

      await expect(_eesee.deploy(
          ESE.address, 
          staking.address,
          proxy.address,
          zeroAddress, 
          accessManager.address,
          zeroAddress,
          zeroAddress,
          zeroAddress
      )).to.be.revertedWithCustomError(_eesee, "InvalidConstructor")

      await expect(_eesee.deploy(
          ESE.address, 
          staking.address,
          proxy.address,
          minter.address,
          zeroAddress,
          zeroAddress, 
          zeroAddress,
          zeroAddress
      )).to.be.revertedWithCustomError(_eesee, "InvalidAccessManager")
    })

    it('Buys tickets without staking contract', async () => {
      const _eesee = await hre.ethers.getContractFactory("EeseeOffchain");

      eesee = await _eesee.deploy(
          ESE.address, 
          zeroAddress,
          proxy.address,
          minter.address, 
          accessManager.address,
          zeroAddress,
          zeroAddress,
          zeroAddress
      )
      await eesee.deployed()

      const asset1 = {token: zeroAddress, tokenID: 1, amount: 1, assetType: 5, data:'0x'}
      const nonce1 = nonce
      const sig1 = await createLot(asset1, "0x01", '999999999999', acc8)
      const _id1 = await eesee.getCreateLotStructHash(asset1, "0x01", nonce1, '999999999999')

      await eesee.connect(signer).createLots(
        [asset1], 
        ["0x01"],
        [sig1]
      )

      await ESE.connect(signer).approve(eesee.address, '10000000000000000000000')
      const sig = await buyTickets(_id1, 500, acc3.address, '999999999999', acc8)
      await expect(eesee.connect(signer).buyTickets([_id1], [500], [sig], acc3.address, "0x")).to.not.be.reverted
  })
});
