const {
    time,
    loadFixture,
  } = require("@nomicfoundation/hardhat-network-helpers");
  const { anyValue } = require("@nomicfoundation/hardhat-chai-matchers/withArgs");
  const { expect } = require("chai");
  const { ethers, network } = require("hardhat");
  const assert = require("assert");
  const { StandardMerkleTree } = require('@openzeppelin/merkle-tree');
  const { getContractAddress } = require('@ethersproject/address')
  const createPermit = require('./utils/createPermit');
  ((network.name != 'multipleNetworks') ? describe : describe.skip)("eeseeDrops", function () {
    let ESE;
    let ERC20;
    let signer, acc2, acc3, acc4, acc5, acc6, acc7, acc8, acc9, feeCollector;
    let ticketBuyers;
    let minter
    let staking
    let eeseeDrops
    let accessManager

    const zeroAddress = "0x0000000000000000000000000000000000000000"
    const oneAddress = "0x0000000000000000000000000000000000000001"
    const getProof = (tree, address) => {
        let proof = null
        for (const [i, v] of tree.entries()) {
            if (v[0] === address) {
                proof = tree.getProof(i);
              }
        }
        return proof
    }
    this.beforeAll(async() => {
        [signer, acc2, acc3, acc4, acc5, acc6, acc7, acc8, acc9, feeCollector, royaltyCollector] = await ethers.getSigners()
        ticketBuyers = [acc2,acc3, acc4, acc5, acc6,  acc7]
        const _AssetTransfer = await hre.ethers.getContractFactory("AssetTransfer");
        const assetTransfer = await _AssetTransfer.deploy()
        await assetTransfer.deployed()


        const _ESE = await hre.ethers.getContractFactory("ESE");
        const _MockERC20 = await hre.ethers.getContractFactory("MockERC20");
        const _mockVRF = await hre.ethers.getContractFactory("MockVRFCoordinator");
        const _eesee = await hre.ethers.getContractFactory("Eesee", {libraries: { AssetTransfer: assetTransfer.address }});
        const _eeseeRandom = await hre.ethers.getContractFactory("EeseeRandomChainlink");
        const _NFT = await hre.ethers.getContractFactory("EeseeNFT");
        const _NFTDrop = await hre.ethers.getContractFactory("EeseeNFTDrop");
        const _NFTlazyMint = await hre.ethers.getContractFactory("EeseeNFTLazyMint");
        const _minter = await hre.ethers.getContractFactory("EeseeMinter");
        const _eeseeDrops = await hre.ethers.getContractFactory("EeseeDrops")
        const _eeseeAccessManager = await hre.ethers.getContractFactory("EeseeAccessManager");
        const _eeseeStaking = await hre.ethers.getContractFactory("EeseeStaking");
        const _mock1InchExecutor = await hre.ethers.getContractFactory("Mock1InchExecutor");
        const _mock1InchRouter = await hre.ethers.getContractFactory("Mock1InchRouter");
        const _eesee1inch = await hre.ethers.getContractFactory("EeseePeripheryOneInch")

        mock1InchRouter = await _mock1InchRouter.deploy();
        await mock1InchRouter.deployed()
        accessManager = await _eeseeAccessManager.deploy();
        await accessManager.deployed()

        ESE = await _ESE.deploy([{
                cliff: 0,
                duration: 0,
                TGEMintShare: 10000
            }
        ])

        await ESE.deployed()
        await ESE.addVestingBeneficiaries(0, [{addr: signer.address, amount: '2000000000000000000000000'}])
        await ESE.initialize(0)

        ERC20 = await _MockERC20.deploy('20000000000000000000000000000')
        await ERC20.deployed()

        mockVRF = await _mockVRF.deploy()
        await mockVRF.deployed()

        mock1InchExecutor = await _mock1InchExecutor.deploy(ESE.address);
        await mock1InchExecutor.deployed()
        await ESE.transfer(mock1InchExecutor.address, '1000000000000000000000000')
        await ERC20.transfer(mock1InchExecutor.address, '1000000000000000000000000')
        await ERC20.transfer(acc3.address, '1000000000000000000000000')

        let eeseeRandom = await _eeseeRandom.deploy(
            {
                vrfCoordinator: mockVRF.address,
                keyHash: '0x0000000000000000000000000000000000000000000000000000000000000000',
                minimumRequestConfirmations: 1,
                callbackGasLimit: 100000
            },
            accessManager.address
        )
        await eeseeRandom.deployed()
        // REQUEST_RANDOM_ROLE
        await accessManager.grantRole('0xde459b4f900be5c5c23ad44aead99bf85b310f34ced089e5c232217a86ec12c3', signer.address)
        // PAUSER_ROLE
        await accessManager.grantRole('0x65d7a28e3265b37a6474929f336521b332c1681b933f6cb9f3376673440d862a', acc8.address)

        const NFTDrop = await _NFTDrop.deploy(zeroAddress)
        const NFTLazyMint = await _NFTlazyMint.deploy(zeroAddress)

        minter = await _minter.deploy(NFTLazyMint.address, NFTDrop.address, eeseeRandom.address)
        await minter.deployed()

        staking = await _eeseeStaking.deploy(
            ESE.address, 
            [{volumeBreakpoint: 500, rewardRateFlexible: 500000, rewardRateLocked: 500000}], 
            accessManager.address,
            zeroAddress
        )
        await staking.deployed()

        let eesee = await _eesee.deploy(
            ESE.address, 
            staking.address,
            oneAddress,
            minter.address, 
            eeseeRandom.address,
            feeCollector.address, 
            oneAddress, 
            accessManager.address,
            zeroAddress,
            zeroAddress,
            zeroAddress
        )
        await eesee.deployed()

        eeseeDrops = await _eeseeDrops.deploy(
            ESE.address,
            feeCollector.address,
            minter.address,
            staking.address,
            accessManager.address,
            zeroAddress
        )
        await eeseeDrops.deployed()

        eesee1inch = await _eesee1inch.deploy(
            mock1InchRouter.address,
            accessManager.address,
            zeroAddress
        )
        await eesee1inch.deployed()

        await staking.grantVolumeUpdater(eeseeDrops.address)

        for (let i = 0; i < ticketBuyers.length; i++) {
            await ESE.transfer(ticketBuyers[i].address, '10000000000000000000000')
            await ESE.connect(ticketBuyers[i]).approve(eeseeDrops.address, '10000000000000000000000')
        }

        await ESE.transfer(acc8.address, '100000000000000000000')
        await ESE.connect(acc8).approve(eeseeDrops.address, '100000000000000000000')
    })
    let merkleTree
    it('drops', async () => {
        await eeseeDrops.connect(signer).changeFee('1000')
        const fee = await eeseeDrops.fee()
        const eeseeNFTDrop = await hre.ethers.getContractFactory("EeseeNFTDrop")
        const leaves = []
        leaves.push([acc2.address])
        leaves.push([acc3.address])
        leaves.push([acc8.address])
        merkleTree = StandardMerkleTree.of(leaves, ['address'])
        const presaleStages = [
            {
                duration: 86400,
                perAddressMintLimit: 5,
                allowListMerkleRoot: merkleTree.root
            },
            {
                duration: 86400,
                perAddressMintLimit: 6,
                allowListMerkleRoot: merkleTree.root
            }
        ]

        await expect(eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [0, ethers.utils.parseUnits('0.02', 'ether')],
            15,
            (await ethers.provider.getBlock()).timestamp + 86400,
            zeroAddress,
            fee
        ))
        .to.be.revertedWithCustomError(eeseeDrops, "InvalidEarningsCollector")

        await expect(eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [0],
            15,
            (await ethers.provider.getBlock()).timestamp + 86400,
            zeroAddress,
            fee
        ))
        .to.be.revertedWithCustomError(eeseeDrops, "InvalidArrayLengths")

        await expect(eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [0, ethers.utils.parseUnits('0.02', 'ether')],
            15,
            (await ethers.provider.getBlock()).timestamp + 86400,
            acc8.address,
            0
        ))
        .to.be.revertedWithCustomError(eeseeDrops, "InvalidFee")

        await expect(eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [0, ethers.utils.parseUnits('0.02', 'ether')],
            15,
            (await ethers.provider.getBlock()).timestamp + 86400,
            acc8.address,
            fee
        ))
        .to.emit(eeseeDrops, "ListDrop").withArgs(0, anyValue, acc8.address)
        .to.emit(minter, "DeployDrop").withArgs(anyValue, eeseeDrops.address)


        const ID = 0
        const listing = await eeseeDrops.drops(ID);
        assert.equal((await eeseeDrops.getMintPrices(ID))[0].toString(), '0', "getMintPrices is correct")
        assert.equal((await eeseeDrops.getMintPrices(ID))[1].toString(), ethers.utils.parseUnits('0.02', 'ether'), "getMintPrices is correct")
        assert.equal(listing.earningsCollector, acc8.address, "earningsCollector is correct")
        assert.equal(listing.fee.toString(), (await eeseeDrops.fee()).toString(), "Fee is correct")

        assert.equal(await eeseeDrops.getDropsLength(), 1, "Length is correct")

        await expect(eeseeDrops.connect(acc2).mintDrops(
            [0], [2], [getProof(merkleTree, acc2.address)], zeroAddress,'0x'
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidRecipient")

        await expect(eeseeDrops.connect(acc2).mintDrops(
            [0], [2,2], [getProof(merkleTree, acc2.address)], acc2.address,'0x'
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidArrayLengths")

        await expect(eeseeDrops.connect(acc2).mintDrops(
            [0], [2], [], acc2.address,'0x'
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidArrayLengths")

        await expect(eeseeDrops.connect(acc2).mintDrops(
            [0], [0], [getProof(merkleTree, acc2.address)], acc2.address,'0x'
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidQuantity")

        await expect(eeseeDrops.connect(acc2).mintDrops(
            [0], [2], [getProof(merkleTree, acc2.address)], acc2.address,'0x'
        )).to.be.revertedWithCustomError(eeseeNFTDrop, "MintingNotStarted")

        await expect(eeseeDrops.connect(acc3).mintDrops(
            [100], [2], [getProof(merkleTree, acc3.address)], acc3.address,'0x'
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidDropID")


        // Presale 1
        await time.increase(86401)

        let balanceBefore = await ESE.balanceOf(acc2.address)
        let collectorBalanceBefore = await ESE.balanceOf(acc8.address)
        let feeBalanceBefore = await ESE.balanceOf(feeCollector.address)
        await expect(eeseeDrops.connect(acc3).mintDrops(
            [0], [2], [getProof(merkleTree, acc3.address)], acc3.address,'0x'
        )).to.emit(eeseeDrops, "MintDrop").withArgs(0, anyValue, acc3.address, 1, 2, 0)
        let balanceAfter = await ESE.balanceOf(acc2.address)
        let collectorBalanceAfter = await ESE.balanceOf(acc8.address)
        let feeBalanceAfter = await ESE.balanceOf(feeCollector.address)

        if(network.name != 'testnet') {
            // Hardhat node has trouble with try catch statements, so we exclude this check for 'testnet'
            // I tested in on the live goerli chain and it was OK.
            assert.equal(await staking.volume(acc2.address), 0, "Volume is correct")
        }
        assert.equal(BigInt(balanceBefore) - BigInt(balanceAfter), 0, "Price paid is correct")
        assert.equal(BigInt(collectorBalanceBefore) - BigInt(collectorBalanceAfter), 0, "Amount collected is correct")
        assert.equal(BigInt(feeBalanceBefore) - BigInt(feeBalanceAfter), 0, "Fee is correct")

        const invalidMerkleTree = StandardMerkleTree.of([[acc4.address]], ['address'])
        await expect(eeseeDrops.connect(acc4).mintDrops(
            [0], [2], [getProof(invalidMerkleTree, acc4.address)], acc4.address,'0x'
        )).to.be.revertedWithCustomError(eeseeNFTDrop, "InvalidProof")

        // Presale 2
        await time.increase(86401)
        let expectedFee = BigInt(ethers.utils.parseUnits('0.04', 'ether')) * BigInt(listing.fee) / BigInt('10000')

        balanceBefore = await ESE.balanceOf(acc2.address)
        collectorBalanceBefore = await ESE.balanceOf(acc8.address)
        feeBalanceBefore = await ESE.balanceOf(feeCollector.address)
        await expect(eeseeDrops.connect(acc2).mintDrops(
            [0], [2], [getProof(merkleTree, acc2.address)], acc2.address,'0x'
        ))
        .to.emit(eeseeDrops, "CollectFee").withArgs(feeCollector.address, expectedFee)
        .and.to.emit(eeseeDrops, "MintDrop").withArgs(0, anyValue, acc2.address, 3, 2, ethers.utils.parseUnits('0.04', 'ether'))
        
        if(network.name != 'testnet') {
            // Hardhat node has trouble with try catch statements, so we exclude this check for 'testnet'
            // I tested in on the live goerli chain and it was OK.
            assert.equal((await staking.volume(acc2.address)).toString(), ethers.utils.parseUnits('0.04', 'ether').toString(), "Volume is correct")
        }
        balanceAfter = await ESE.balanceOf(acc2.address)
        collectorBalanceAfter = await ESE.balanceOf(acc8.address)
        feeBalanceAfter = await ESE.balanceOf(feeCollector.address)

        assert.equal(BigInt(balanceBefore) - BigInt(balanceAfter), ethers.utils.parseUnits('0.04', 'ether'), "Price paid is correct")
        assert.equal(BigInt(feeBalanceAfter) - BigInt(feeBalanceBefore), expectedFee, "Fee is correct")

        let expectedReceive = BigInt(ethers.utils.parseUnits('0.04', 'ether')) - expectedFee
        assert.equal(BigInt(collectorBalanceAfter) - BigInt(collectorBalanceBefore), expectedReceive, "Amount collected is correct")

        await expect(eeseeDrops.connect(acc4).mintDrops(
            [0], [2], [getProof(merkleTree, acc2.address)], acc4.address,'0x'
        )).to.be.revertedWithCustomError(eeseeNFTDrop, "InvalidProof")
    })
    it('mint drop with swap', async () => {

        const ID = 0
        const listing = await eeseeDrops.drops(ID);
        //const merkleTree = StandardMerkleTree.of([[acc3.address]], ['address'])
        expectedFee = BigInt(ethers.utils.parseUnits('0.04', 'ether')) * BigInt(listing.fee) / BigInt('10000')
        balanceBefore = await ESE.balanceOf(acc3.address)
        collectorBalanceBefore = await ESE.balanceOf(acc8.address)
        feeBalanceBefore = await ESE.balanceOf(feeCollector.address)

        const interface = new ethers.utils.Interface([
            'function swap(address executor, tuple(address srcToken, address dstToken, address srcReceiver, address dstReceiver, uint amount, uint minReturnAmount, uint flags) desc, bytes permit, bytes data)',
            'function swep(address executor, tuple(address srcToken, address dstToken, address srcReceiver, address dstReceiver, uint amount, uint minReturnAmount, uint flags) desc, bytes permit, bytes data)'
        ]);
        let swapAmount = ethers.utils.parseUnits('0.03', 'ether').add(11) 

        const snapshotId = await network.provider.send('evm_snapshot')
        await ESE.connect(acc3).approve(eesee1inch.address, swapAmount.mul('2'))
        let swapData = interface.encodeFunctionData('swap', [
            mock1InchExecutor.address, 
            {
                srcToken: ESE.address,//
                dstToken: ESE.address,
                srcReceiver: mock1InchExecutor.address, 
                dstReceiver: eesee1inch.address, 
                amount: swapAmount.mul('2'), 
                minReturnAmount: 100,
                flags: 1,
            }, 
            '0x00',
            '0x000000000000000000000000' + ERC20.address.substring(2)
        ])

        let mintDropsData = eeseeDrops.interface.encodeFunctionData('mintDrops', [
            [0], [2], [getProof(merkleTree, acc3.address)],
            acc3.address,
            "0x"
        ])
        await expect(eesee1inch.connect(acc3).executeWithSwap(swapData, eeseeDrops.address, mintDropsData))
        .to.not.be.reverted
        await network.provider.send("evm_revert", [snapshotId])

        swapData = interface.encodeFunctionData('swap', [
            mock1InchExecutor.address, 
            {
                srcToken: zeroAddress,//
                dstToken: ERC20.address,
                srcReceiver: mock1InchExecutor.address, 
                dstReceiver: eesee1inch.address, 
                amount: 0, 
                minReturnAmount: 100,
                flags: 1,
            }, 
            '0x00',
            '0x000000000000000000000000' + ERC20.address.substring(2)
        ])
        await expect(eesee1inch.connect(acc3).executeWithSwap(swapData, eeseeDrops.address, mintDropsData, {value: swapAmount}))
        .to.be.revertedWithCustomError(eesee1inch, "InvalidAmount")

        swapAmount = ethers.utils.parseUnits('0.03', 'ether').add(11) 
        swapData = interface.encodeFunctionData('swap', [
            mock1InchExecutor.address, 
            {
                srcToken: zeroAddress,//
                dstToken: ESE.address,
                srcReceiver: mock1InchExecutor.address, 
                dstReceiver: eesee1inch.address, 
                amount: swapAmount, 
                minReturnAmount: 100,
                flags: 1,
            }, 
            '0x00',
            '0x000000000000000000000000' + ERC20.address.substring(2)
        ])
        await expect(eesee1inch.connect(acc3).executeWithSwap(swapData, eeseeDrops.address, mintDropsData, {value: 1}))
        .to.be.revertedWithCustomError(eesee1inch, "InvalidMsgValue")

        swapAmount = ethers.utils.parseUnits('0.06', 'ether')
        await ERC20.connect(acc3).approve(eesee1inch.address, swapAmount)
        
        swapData = interface.encodeFunctionData('swap', [
            mock1InchExecutor.address, 
            {
                srcToken: ERC20.address,//
                dstToken: ESE.address,
                srcReceiver: mock1InchExecutor.address, 
                dstReceiver: eesee1inch.address, 
                amount: swapAmount, 
                minReturnAmount: 100,
                flags: 1,
            }, 
            '0x00',
            '0x000000000000000000000000' + ERC20.address.substring(2)
        ])
        await expect(eesee1inch.connect(acc3).executeWithSwap(swapData, eeseeDrops.address, mintDropsData))
        .to.be.revertedWithCustomError(eesee1inch, "ExecutionFailed")

        swapAmount = ethers.utils.parseUnits('0.03', 'ether').add(11) 
        swapData = interface.encodeFunctionData('swap', [
            mock1InchExecutor.address, 
            {
                srcToken: zeroAddress,//
                dstToken: ESE.address,
                srcReceiver: mock1InchExecutor.address, 
                dstReceiver: eesee1inch.address, 
                amount: swapAmount, //amount - 10 * 2
                minReturnAmount: 100,
                flags: 1,
            }, 
            '0x00',
            '0x000000000000000000000000' + ERC20.address.substring(2)
        ])

        await expect(eesee1inch.connect(signer).pause()).to.be.revertedWithCustomError(eesee1inch, "CallerNotAuthorized")
        await expect(eesee1inch.connect(acc8).pause()).to.emit(eesee1inch, "Paused").withArgs(acc8.address)
        await expect(eesee1inch.connect(acc3).executeWithSwap(swapData, eeseeDrops.address, mintDropsData, {value: swapAmount}))
            .to.be.revertedWith("Pausable: paused")
        await expect(eesee1inch.connect(signer).unpause()).to.be.revertedWithCustomError(eesee1inch, "CallerNotAuthorized")
        await expect(eesee1inch.connect(acc8).unpause()).to.emit(eesee1inch, "Unpaused").withArgs(acc8.address)

        await expect(eesee1inch.connect(acc3).executeWithSwap(swapData, eeseeDrops.address, mintDropsData, {value: swapAmount}))
        .to.emit(eeseeDrops, "MintDrop").withArgs(0, anyValue, acc3.address, 5, 2, ethers.utils.parseUnits('0.04', 'ether'))
        .and.to.emit(eeseeDrops, "CollectFee").withArgs(feeCollector.address, expectedFee)
        
        balanceAfter = await ESE.balanceOf(acc3.address)
        collectorBalanceAfter = await ESE.balanceOf(acc8.address)
        feeBalanceAfter = await ESE.balanceOf(feeCollector.address)

        if(network.name != 'testnet') {
            // Hardhat node has trouble with try catch statements, so we exclude this check for 'testnet'
            // I tested in on the live goerli chain and it was OK.
            assert.equal((await staking.volume(acc2.address)).toString(), ethers.utils.parseUnits('0.04', 'ether').toString(), "Volume is correct")
        }

        assert.equal(BigInt(balanceAfter) - BigInt(balanceBefore), '20000000000000001', "Dust is correct")
        assert.equal(BigInt(feeBalanceAfter) - BigInt(feeBalanceBefore), expectedFee, "Fee is correct")

        expectedReceive = BigInt(ethers.utils.parseUnits('0.04', 'ether')) - expectedFee
        assert.equal(BigInt(collectorBalanceAfter) - BigInt(collectorBalanceBefore), expectedReceive, "Amount collected is correct")


        balanceBefore = await ESE.balanceOf(acc3.address)
        collectorBalanceBefore = await ESE.balanceOf(acc8.address)
        feeBalanceBefore = await ESE.balanceOf(feeCollector.address)

        swapAmount = ethers.utils.parseUnits('0.12', 'ether').add(12)
        await ERC20.connect(acc3).approve(eesee1inch.address, swapAmount)
        
        swapData = interface.encodeFunctionData('swap', [
            mock1InchExecutor.address, 
            {
                srcToken: ERC20.address,//
                dstToken: ESE.address,
                srcReceiver: mock1InchExecutor.address, 
                dstReceiver: eesee1inch.address, 
                amount: swapAmount, 
                minReturnAmount: 100,
                flags: 1,
            }, 
            '0x00',
            '0x000000000000000000000000' + ERC20.address.substring(2)
        ])
        await expect(eesee1inch.connect(acc3).executeWithSwap(swapData, eeseeDrops.address, mintDropsData))
        .to.emit(eeseeDrops, "MintDrop").withArgs(0, anyValue, acc3.address, 7, 2, ethers.utils.parseUnits('0.04', 'ether'))
        .and.to.emit(eeseeDrops, "CollectFee").withArgs(feeCollector.address, expectedFee)

        balanceAfter = await ESE.balanceOf(acc3.address)
        collectorBalanceAfter = await ESE.balanceOf(acc8.address)
        feeBalanceAfter = await ESE.balanceOf(feeCollector.address)

        if(network.name != 'testnet') {
            // Hardhat node has trouble with try catch statements, so we exclude this check for 'testnet'
            // I tested in on the live goerli chain and it was OK.
            assert.equal((await staking.volume(acc2.address)).toString(), ethers.utils.parseUnits('0.04', 'ether').toString(), "Volume is correct")
        }
        assert.equal( BigInt(balanceAfter) - BigInt(balanceBefore), '20000000000000001', "Dust is correct")
        assert.equal(BigInt(feeBalanceAfter) - BigInt(feeBalanceBefore), expectedFee, "Fee is correct")

        expectedReceive = BigInt(ethers.utils.parseUnits('0.04', 'ether')) - expectedFee
        assert.equal(BigInt(collectorBalanceAfter) - BigInt(collectorBalanceBefore), expectedReceive, "Amount collected is correct")
    })
    it('mint drops with permit',  async () => {
        const ID = 0
        const listing = await eeseeDrops.drops(ID);
        //const merkleTree = StandardMerkleTree.of([[acc4.address]], ['address'])
        expectedFee = BigInt(ethers.utils.parseUnits('0.04', 'ether')) * BigInt(listing.fee) / BigInt('10000')
        balanceBefore = await ESE.balanceOf(acc3.address)
        collectorBalanceBefore = await ESE.balanceOf(acc8.address)
        feeBalanceBefore = await ESE.balanceOf(feeCollector.address)
        const currentTimestamp = (await ethers.provider.getBlock()).timestamp
        const deadline = currentTimestamp + 10000
        const incorrectPermit = await createPermit(acc3, eeseeDrops, '10', deadline, ESE)
        const params_ = ethers.utils.defaultAbiCoder.encode(
            ["uint256", "uint256", "uint8", "bytes32", "bytes32"],
            [ethers.utils.parseUnits('0.04', 'ether'), deadline, incorrectPermit.v, incorrectPermit.r, incorrectPermit.s]
          );

        await expect(eeseeDrops.connect(acc3).mintDrops(
            [0], [2], [getProof(merkleTree, acc3.address)], acc3.address, params_
        )).to.be.revertedWith("ERC20Permit: invalid signature")

        const params_2 = ethers.utils.defaultAbiCoder.encode(
            ["uint256", "uint256", "uint8", "bytes32", "bytes32"],
            ['10', deadline, incorrectPermit.v, incorrectPermit.r, incorrectPermit.s]
          );
        await expect(eeseeDrops.connect(acc3).mintDrops(
            [0], [2], [getProof(merkleTree, acc3.address)], acc3.address, params_2
        )).to.be.revertedWith("ERC20: insufficient allowance")

        await staking.revokeVolumeUpdater(eeseeDrops.address) 
        
        const volumeBefore = await staking.volume(acc3.address)

        const permit = await createPermit(acc3, eeseeDrops, ethers.utils.parseUnits('0.04', 'ether'), deadline, ESE)
        const params_3 = ethers.utils.defaultAbiCoder.encode(
            ["uint256", "uint256", "uint8", "bytes32", "bytes32"],
            [ethers.utils.parseUnits('0.04', 'ether'), deadline, permit.v, permit.r, permit.s]
          );
        await expect(eeseeDrops.connect(acc3).mintDrops(
            [0], [2], [getProof(merkleTree, acc3.address)], acc3.address, params_3
        ))
        .to.emit(eeseeDrops, "MintDrop").withArgs(0, anyValue, acc3.address, 9, 2, ethers.utils.parseUnits('0.04', 'ether'))
        .and.to.emit(eeseeDrops, "CollectFee").withArgs(feeCollector.address, expectedFee)
        .and.to.emit(eeseeDrops, "AddVolumeReverted").withArgs(ethers.utils.parseUnits('0.04', 'ether'), acc3.address)
        
        balanceAfter = await ESE.balanceOf(acc3.address)
        collectorBalanceAfter = await ESE.balanceOf(acc8.address)
        feeBalanceAfter = await ESE.balanceOf(feeCollector.address)

        if(network.name != 'testnet') {
            // Hardhat node has trouble with try catch statements, so we exclude this check for 'testnet'
            // I tested in on the live goerli chain and it was OK.
            const volumeAfter = await staking.volume(acc3.address)
            assert.equal(volumeAfter.sub(volumeBefore).toString(), '0', "Volume is correct")
        }
        assert.equal(BigInt(balanceBefore) - BigInt(balanceAfter), ethers.utils.parseUnits('0.04', 'ether'), "Price paid is correct")
        assert.equal(BigInt(feeBalanceAfter) - BigInt(feeBalanceBefore), expectedFee, "Fee is correct")

        expectedReceive = BigInt(ethers.utils.parseUnits('0.04', 'ether')) - expectedFee
        assert.equal(BigInt(collectorBalanceAfter) - BigInt(collectorBalanceBefore), expectedReceive, "Amount collected is correct")
        // Reapprove after permit
        await ESE.connect(acc3).approve(eeseeDrops.address, '100000000000000000000')
    })
    it('mint drops with different fees', async () => {
        const snapshotId = await network.provider.send('evm_snapshot')
        const leaves = []
        leaves.push([acc2.address])
        leaves.push([acc3.address])
        leaves.push([acc8.address])
        merkleTree = StandardMerkleTree.of(leaves, ['address'])
        const presaleStages = [
            {
                duration: 86400,
                perAddressMintLimit: 5,
                allowListMerkleRoot: merkleTree.root
            },
            {
                duration: 86400,
                perAddressMintLimit: 6,
                allowListMerkleRoot: merkleTree.root
            }
        ]
        await eeseeDrops.connect(signer).changeFee('5000')
        let fee = await eeseeDrops.fee()
        await expect(eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [ethers.utils.parseUnits('0.02', 'ether'), ethers.utils.parseUnits('0.02', 'ether')],
            15,
            (await ethers.provider.getBlock()).timestamp + 86400,
            acc8.address,
            fee
        ))
        .to.emit(eeseeDrops, "ListDrop")
        .withArgs(1, anyValue, acc8.address)
        
        await time.increase(86401)
        await expect(eeseeDrops.connect(acc2).mintDrops(
            [1], [2], [getProof(merkleTree, acc2.address)], acc2.address,'0x'
        ))
        .to.emit(eeseeDrops, "CollectFee").withArgs(feeCollector.address, ethers.utils.parseUnits('0.02', 'ether'))
        .and.to.emit(eeseeDrops, "MintDrop").withArgs(1, anyValue, acc2.address, 1, 2, ethers.utils.parseUnits('0.04', 'ether'))


        await eeseeDrops.connect(signer).changeFee(0)
        fee = await eeseeDrops.fee()
        await expect(eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [ethers.utils.parseUnits('0.02', 'ether'), ethers.utils.parseUnits('0.02', 'ether')],
            15,
            (await ethers.provider.getBlock()).timestamp + 86400,
            acc8.address,
            fee
        ))
        .to.emit(eeseeDrops, "ListDrop")
        .withArgs(2, anyValue, acc8.address)

        await time.increase(86401)
        await expect(eeseeDrops.connect(acc2).mintDrops(
            [2], [2], [getProof(merkleTree, acc2.address)], acc2.address,'0x'
        ))
        .and.to.emit(eeseeDrops, "MintDrop").withArgs(2, anyValue, acc2.address, 1, 2, ethers.utils.parseUnits('0.04', 'ether'))


        await eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            [presaleStages[0]],
            ['0'],
            15,
            (await ethers.provider.getBlock()).timestamp + 86400,
            acc8.address,
            fee
        )
        await time.increase(86401)
        await expect(eeseeDrops.connect(acc2).mintDrops(
            [3], [2], [getProof(merkleTree, acc2.address)], acc2.address,'0x'
        ))
        .to.emit(eeseeDrops, "MintDrop").withArgs(3, anyValue, acc2.address, 1, 2, 0)
        await network.provider.send("evm_revert", [snapshotId])
    })

    it('is nonReentrant', async () => {
        const snapshotId = await network.provider.send('evm_snapshot')
        const _eeseeReentrancy = await hre.ethers.getContractFactory("EeseeDropsReentrancy");
        eeseeReentrancy = await _eeseeReentrancy.deploy(eeseeDrops.address);
        await eeseeReentrancy.deployed()

        fee = await eeseeDrops.fee()
        const leaves = []
        leaves.push([eeseeReentrancy.address])
        merkleTree = StandardMerkleTree.of(leaves, ['address'])
        const presaleStages = [{
            duration: 86400,
            perAddressMintLimit: 0,
            allowListMerkleRoot: merkleTree.root
        }]

        await eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [0],
            10,
            (await ethers.provider.getBlock()).timestamp + 5,
            acc8.address,
            fee
        )
        await time.increase(10)


        await expect(eeseeDrops.connect(acc4).mintDrops(
            [1], 
            [1], 
            [getProof(merkleTree, eeseeReentrancy.address)], 
            eeseeReentrancy.address,
            '0x'
        )).to.be.revertedWith("ReentrancyGuard: reentrant call")

        await network.provider.send("evm_revert", [snapshotId])
    })

    it('buys multiple drops', async () => {
        fee = await eeseeDrops.fee()
        const leaves = []
        leaves.push([acc2.address])
        leaves.push([acc3.address])
        leaves.push([acc8.address])
        merkleTree = StandardMerkleTree.of(leaves, ['address'])
        const presaleStages = [{
            name: 'Public Stage',
            mintFee: ethers.utils.parseUnits('0.01', 'ether'),
            duration: 86400,
            perAddressMintLimit: 0,
            allowListMerkleRoot: '0x0000000000000000000000000000000000000000000000000000000000000000'
        }]

        await eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [ethers.utils.parseUnits('0.01', 'ether')],
            10,
            (await ethers.provider.getBlock()).timestamp + 5,
            acc8.address,
            fee
        )
        await eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                revealedURI: "/newURL",
                baseURI: "/",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [ethers.utils.parseUnits('0.01', 'ether')],
            10,
            (await ethers.provider.getBlock()).timestamp + 5,
            acc8.address,
            fee
        )

        await time.increase(10)


        expectedFee = BigInt(ethers.utils.parseUnits('0.03', 'ether')) * BigInt('1000') / BigInt('10000')
        await expect(await eeseeDrops.connect(acc4).mintDrops(
            [1,2], [1,2], [[],[]], acc4.address,'0x'
        ))
        .to.emit(eeseeDrops, "MintDrop").withArgs(1, anyValue, acc4.address, 1, 1, ethers.utils.parseUnits('0.01', 'ether'))
        .to.emit(eeseeDrops, "MintDrop").withArgs(2, anyValue, acc4.address, 1, 2, ethers.utils.parseUnits('0.02', 'ether'))
        .and.to.emit(eeseeDrops, "CollectFee").withArgs(feeCollector.address, expectedFee)
    })

    it('reverts constructor', async () => {
        const EeseeDropsFactory = await hre.ethers.getContractFactory('EeseeDrops')
        await expect(EeseeDropsFactory.deploy(
            zeroAddress,
            feeCollector.address,
            minter.address,
            staking.address,
            accessManager.address,
            oneAddress
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidConstructor")
        await expect(EeseeDropsFactory.deploy(
            ESE.address,
            zeroAddress,
            minter.address,
            staking.address,
            accessManager.address,
            oneAddress
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidConstructor")
        await expect(EeseeDropsFactory.deploy(
            ESE.address,
            feeCollector.address,
            zeroAddress,
            staking.address,
            accessManager.address,
            oneAddress
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidConstructor")
        await expect(EeseeDropsFactory.deploy(
            ESE.address,
            feeCollector.address,
            minter.address,
            staking.address,
            zeroAddress,
            oneAddress
        )).to.be.revertedWithCustomError(eeseeDrops, "InvalidAccessManager")
    })

    it('Mints drops without staking contract', async () => {
        const _eeseeDrops = await hre.ethers.getContractFactory("EeseeDrops")
        eeseeDrops = await _eeseeDrops.deploy(
            ESE.address,
            feeCollector.address,
            minter.address,
            zeroAddress,
            accessManager.address,
            oneAddress
        )
        await eeseeDrops.deployed()

        
        fee = await eeseeDrops.fee()
        const leaves = []
        leaves.push([acc2.address])
        leaves.push([acc3.address])
        leaves.push([acc8.address])
        merkleTree = StandardMerkleTree.of(leaves, ['address'])
        const presaleStages = [{
            name: 'Public Stage',
            mintFee: ethers.utils.parseUnits('0.01', 'ether'),
            duration: 86400,
            perAddressMintLimit: 0,
            allowListMerkleRoot: '0x0000000000000000000000000000000000000000000000000000000000000000'
        }]

        await eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                baseURI: "/",
                revealedURI: "/newURL",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [ethers.utils.parseUnits('0.01', 'ether')],
            10,
            (await ethers.provider.getBlock()).timestamp + 5,
            acc8.address,
            fee
        )
        await eeseeDrops.connect(acc8).listDrop(
            {
                name: "APES",
                symbol:"bayc",
                revealedURI: "/newURL",
                baseURI: "/",
                contractURI:'/',
                royaltyReceiver: acc8.address,
                royaltyFeeNumerator: 300
            },
            presaleStages,
            [ethers.utils.parseUnits('0.01', 'ether')],
            10,
            (await ethers.provider.getBlock()).timestamp + 5,
            acc8.address,
            fee
        )

        await time.increase(10)

        await ESE.connect(acc4).approve(eeseeDrops.address, '100000000000000000000000000000000')


        expectedFee = BigInt(ethers.utils.parseUnits('0.03', 'ether')) * BigInt('1000') / BigInt('10000')
        await expect(await eeseeDrops.connect(acc4).mintDrops(
            [0,1], [1,2], [[],[]], acc4.address,'0x'
        )).to.not.be.reverted
    })

    it('admin', async () => {
        let newValue = 1
        const fee = await eeseeDrops.fee() 
        await expect(eeseeDrops.connect(acc2).changeFee(newValue)).to.be.revertedWithCustomError(eeseeDrops, "CallerNotAuthorized")
        await expect(eeseeDrops.connect(signer).changeFee('5001')).to.be.revertedWithCustomError(eeseeDrops, "FeeTooHigh")
        await expect(eeseeDrops.connect(signer).changeFee(newValue))
        .to.emit(eeseeDrops, "ChangeFee")
        .withArgs(fee, newValue)
        assert.equal(newValue, await eeseeDrops.fee(), "fee has changed")

        await expect(eeseeDrops.connect(acc2).changeFeeSplitter(oneAddress)).to.be.revertedWithCustomError(eeseeDrops, "CallerNotAuthorized")
        await expect(eeseeDrops.connect(signer).changeFeeSplitter(zeroAddress)).to.be.revertedWithCustomError(eeseeDrops, "InvalidFeeSplitter")
        await expect(eeseeDrops.connect(signer).changeFeeSplitter(oneAddress))
        .to.emit(eeseeDrops, "ChangeFeeSplitter")
        .withArgs(feeCollector.address, oneAddress)
        assert.equal(oneAddress, await eeseeDrops.feeSplitter(), "feesplitter has changed")
    })
})